-- Filename:     ad7813_module_sim.do
-- Filetype:     VHDL Source Code
-- Date:         23 sep 2012
-- Update:       -
-- Description:  VHDL Description of a AD7813 ADC Converter
-- Author:       J. op den Brouw
-- State:        Demo
-- Error:        -
-- Version:      1.0
-- Copyright:    (c)2012, De Haagse Hogeschool

-- This file contains a VHDL description of an AD7813 ADC converter, see
-- ( http://www.analog.com/en/analog-to-digital-converters/ad-converters/ad7813/products/product.html )
-- for more information.
-- The VHDL description can both be simulated and synthesized.
--
-- Please not that this design opposite to the real design, has a clock and reset
-- input. Please supply a clock frequency of > 10 MHz. The reset is active HIGH.
-- This design has a 10 bit "analog" data input that represents the analog input.
--
-- Undocumented: this design treats the CS signal as a kind of toggle signal.
-- When first applied, the data readout is the higher 8 bit of the 10 bit data.
-- When secondly applied, the data readout is the low 2 bit, concatenated with
-- all zeros, of the 10 bit data. When applied further more, the readout toggles
-- between the first and second situation.
-- When a readout action is done during BUSY high, the readout circuitry is reset,
-- and the databus is all High-Z.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity ad7813_module_sim is
	port (clk      : in std_logic;
			areset   : in std_logic;
			analog   : in std_logic_vector(9 downto 0);
			CONVST   : in std_logic;
			CS       : in std_logic;
			RD       : in std_logic;
			BUSY     : out std_logic;
			DB       : out std_logic_vector(7 downto 0)
		);
end entity ad7813_module_sim;

architecture rtl of ad7813_module_sim is
signal counter23us : integer range 0 to 114;
signal counter15us : integer range 0 to 74;
-- ADC is 10 bits!
signal db_int : std_logic_vector(9 downto 0);
signal busy_int : std_logic;
signal convst_int : std_logic;

type conversion_state_type is (idle, wait_for_start, wait_convst_start, check_mode, mode_1_op, mode_2_op,
							mode_1_op_busy, mode_2_op_busy, busy_st, busy_downcount);
signal conversion_state : conversion_state_type;

type timer_state_type is (idle, wait_for_start, countdown, wait_for_neg);
signal timer_state : timer_state_type;

type dataread_state_type is (idle, wait_for_cs_low, readout_msb, wait_readout_lsb, readout_lsb);
signal dataread_state : dataread_state_type;

begin

	-- Busy signal output
	BUSY <= busy_int;
	
	-- The 1.5 us internal trigger timer
	-- See datasheet pg 8
	process (clk, areset) is
	begin
		if areset = '1' then
			timer_state <= idle;
			counter15us <= 0;
			convst_int <= '0';
		elsif rising_edge(clk) then
			case timer_state is
				-- Idle mode, first state after reset
				when idle =>
					timer_state <= wait_for_start;
					convst_int <= '0';
				-- Wait for conversion start
				when wait_for_start =>
					convst_int <= '0';
					counter15us <= 74;
					if CONVST = '1' then
						timer_state <= countdown;
					end if;
				-- A conversion start is detected, so downcount...
				-- This will create a delay of 1.5 us, and sets
				-- the internal conversion start.
				when countdown =>
					convst_int <= '1';
					if counter15us = 0 then
						timer_state <= wait_for_neg;
					else
						counter15us <= counter15us - 1;
					end if;
				-- Wait for conversion signal to negate, as not to
				-- start a new countdown.
				when wait_for_neg =>
					convst_int <= '0';
					if CONVST = '0' then
						timer_state <= idle;
					end if;
				when others =>
					null;
			end case;
		end if;
	end process;

   -- The conversion process	
	process (clk, areset) is
	begin
		if areset = '1' then
			conversion_state <= idle;
			busy_int <= '0';
		elsif rising_edge(clk) then
			case conversion_state is
				-- Idle state, when wakeup from reset
				when idle =>
					busy_int <= '0';
					-- wait for the convst-signals to be cleared
					if CONVST = '0' and convst_int = '0' then
						conversion_state <= wait_for_start;
					end if;
				when wait_for_start =>
					busy_int <= '0';
					-- start conversion, also trigger the 1.5 us timer
					if CONVST = '1' then
						conversion_state <= wait_convst_start;
					end if;
				when wait_convst_start =>
					busy_int <= '0';
					-- Wait for internal conv to start
					if convst_int = '1' then
						conversion_state <= check_mode;
					end if;
				when check_mode =>
					busy_int <= '0';
					-- Internal timer done and external conv high, then mode 1
					-- as per datasheet pg 8.
					if convst_int = '0' and CONVST = '1' then
						conversion_state <= mode_1_op;
					-- Internal timer done and external conv low, then mode 2
					-- as per datasheet pg 8.
					elsif convst_int = '0' and CONVST = '0' then
						conversion_state <= mode_2_op;
					end if;
				when mode_1_op =>
					-- Mode 1 op, wait for external conv to end
					busy_int <= '0';
					if CONVST = '0' then
						conversion_state <= mode_1_op_busy;
					end if;
				when mode_2_op =>
					-- Mode 2 op, wait for internal conv to end
					busy_int <= '0';
					if convst_int = '0' then
						conversion_state <= mode_2_op_busy;
					end if;
				when mode_1_op_busy | mode_2_op_busy =>
					-- Wait one cycle...
					busy_int <= '0';
					conversion_state <= busy_st;
				when busy_st =>
					-- We're busy converting...
					busy_int <= '1';
					counter23us <= 114;
					conversion_state <= busy_downcount;
				when busy_downcount =>
					-- Timer for converting
					busy_int <= '1';
					if counter23us = 0 then
						conversion_state <= idle;
					else
						counter23us <= counter23us - 1;
					end if;
				when others =>
					null;
			end case;
		end if;
	end process;
	
	-- The Data Readout Process - state machine
	process (clk, areset) is
	begin
		if areset = '1' then
			dataread_state <= idle;
			db_int <= (others => '0');
		elsif rising_edge(clk) then
			-- A busy resets the databus/readout
			if busy_int = '1' then
				dataread_state <= idle;
			else
				case dataread_state is
					when idle =>
						-- We're idle, just jump to the next state
						dataread_state <= wait_for_cs_low;
					when wait_for_cs_low =>
						-- Wait for CS to go low for the first time
						-- This will read out the high 8 bits
						if CS = '0' then
							dataread_state <= readout_msb;
							db_int <= analog;
						end if;
					when readout_msb =>
						-- Wait for CS to go high for the first time.
						if CS = '1'then
							dataread_state <= wait_readout_lsb;
						end if;
					when wait_readout_lsb =>
						-- Wait for CS to go low for the second time.
						-- This will read out the low 2 bits
						if CS = '0'then
							dataread_state <= readout_lsb;
						end if;
					when readout_lsb =>
						-- Wait for CS to go high for the second time.
						if CS = '1'then
							dataread_state <= wait_for_cs_low;
						end if;
					when others =>
						null;
				end case;
			end if;
		end if;
	end process;

	-- The Data Output -  data output
	process (dataread_state, db_int, RD) is
	begin
			-- The default value of the databus is High-Z
			DB <= (others => 'Z');
			case dataread_state is
				when idle =>
					null;
				when wait_for_cs_low =>
					null;
				when readout_msb =>
					-- Reading the msb 8 bit ONLY if RD is active
					if RD = '0' then
						DB <= db_int(9 downto 2);
					end if;
				when wait_readout_lsb =>
					null;
				when readout_lsb =>
					-- Reading the lsb 2 bit ONLY if RD is active
					-- appended with zeros as per datasheet pg 9.
					if RD = '0' then
						DB <= db_int(1 downto 0) & "000000";
					end if;
				when others =>
					null;
			end case;
	end process;
	
end architecture rtl;
